@@ -1,15 +1,16 @@
-/* Copyright (C) 2015-2016 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
+/* Copyright (C) 2015-2017 Jason A. Donenfeld <Jason@zx2c4.com>. All Rights Reserved. */
 
 #include "device.h"
 #include "peer.h"
 #include "socket.h"
-#include "packets.h"
+#include "queueing.h"
 #include "messages.h"
 
 #include <linux/ctype.h>
 #include <linux/net.h>
 #include <linux/if_vlan.h>
 #include <linux/if_ether.h>
+#include <linux/inetdevice.h>
 #include <net/udp_tunnel.h>
 #include <net/ipv6.h>
 
@@ -20,6 +21,7 @@ static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct
 		.daddr = endpoint->addr4.sin_addr.s_addr,
 		.fl4_dport = endpoint->addr4.sin_port,
 		.fl4_sport = htons(wg->incoming_port),
+		.flowi4_mark = wg->fwmark,
 		.flowi4_proto = IPPROTO_UDP
 	};
 	struct rtable *rt = NULL;
@@ -27,10 +29,10 @@ static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct
 	int ret = 0;
 
 	skb->next = skb->prev = NULL;
-	skb->dev = netdev_pub(wg);
+	skb->dev = wg->dev;
 
-	rcu_read_lock();
-	sock = rcu_dereference(wg->sock4);
+	rcu_read_lock_bh();
+	sock = rcu_dereference_bh(wg->sock4);
 
 	if (unlikely(!sock)) {
 		ret = -ENONET;
@@ -42,38 +44,40 @@ static inline int send4(struct wireguard_device *wg, struct sk_buff *skb, struct
 
 	if (!rt) {
 		security_sk_classify_flow(sock, flowi4_to_flowi(&fl));
-		rt = ip_route_output_flow(sock_net(sock), &fl, sock);
-		if (unlikely(IS_ERR(rt) && PTR_ERR(rt) == -EINVAL && fl.saddr)) {
-			endpoint->src4.s_addr = fl.saddr = 0;
+		if (unlikely(!inet_confirm_addr(sock_net(sock), NULL, 0, fl.saddr, RT_SCOPE_HOST))) {
+			endpoint->src4.s_addr = *(__force __be32 *)&endpoint->src_if4 = fl.saddr = 0;
 			if (cache)
 				dst_cache_reset(cache);
+		}
+		rt = ip_route_output_flow(sock_net(sock), &fl, sock);
+		if (unlikely(endpoint->src_if4 && ((IS_ERR(rt) && PTR_ERR(rt) == -EINVAL) || (!IS_ERR(rt) && rt->dst.dev->ifindex != endpoint->src_if4)))) {
+			endpoint->src4.s_addr = *(__force __be32 *)&endpoint->src_if4 = fl.saddr = 0;
+			if (cache)
+				dst_cache_reset(cache);
+			if (!IS_ERR(rt))
+				ip_rt_put(rt);
 			rt = ip_route_output_flow(sock_net(sock), &fl, sock);
 		}
 		if (unlikely(IS_ERR(rt))) {
 			ret = PTR_ERR(rt);
-			net_dbg_ratelimited("No route to %pISpfsc, error %d\n", &endpoint->addr, ret);
+			net_dbg_ratelimited("%s: No route to %pISpfsc, error %d\n", wg->dev->name, &endpoint->addr, ret);
 			goto err;
 		} else if (unlikely(rt->dst.dev == skb->dev)) {
-			dst_release(&rt->dst);
+			ip_rt_put(rt);
 			ret = -ELOOP;
-			net_dbg_ratelimited("Avoiding routing loop to %pISpfsc\n", &endpoint->addr);
+			net_dbg_ratelimited("%s: Avoiding routing loop to %pISpfsc\n", wg->dev->name, &endpoint->addr);
 			goto err;
 		}
 		if (cache)
 			dst_cache_set_ip4(cache, &rt->dst, fl.saddr);
 	}
-
-	udp_tunnel_xmit_skb(rt, sock, skb,
-			    fl.saddr, fl.daddr,
-			    ds, ip4_dst_hoplimit(&rt->dst), 0,
-			    fl.fl4_sport, fl.fl4_dport,
-			    false, false);
+	udp_tunnel_xmit_skb(rt, sock, skb, fl.saddr, fl.daddr, ds, ip4_dst_hoplimit(&rt->dst), 0, fl.fl4_sport, fl.fl4_dport, false, false);
 	goto out;
 
 err:
 	kfree_skb(skb);
 out:
-	rcu_read_unlock();
+	rcu_read_unlock_bh();
 	return ret;
 }
 
@@ -85,6 +89,7 @@ static inline int send6(struct wireguard_device *wg, struct sk_buff *skb, struct
 		.daddr = endpoint->addr6.sin6_addr,
 		.fl6_dport = endpoint->addr6.sin6_port,
 		.fl6_sport = htons(wg->incoming_port),
+		.flowi6_mark = wg->fwmark,
 		.flowi6_oif = endpoint->addr6.sin6_scope_id,
 		.flowi6_proto = IPPROTO_UDP
 		/* TODO: addr->sin6_flowinfo */
@@ -94,10 +99,10 @@ static inline int send6(struct wireguard_device *wg, struct sk_buff *skb, struct
 	int ret = 0;
 
 	skb->next = skb->prev = NULL;
-	skb->dev = netdev_pub(wg);
+	skb->dev = wg->dev;
 
-	rcu_read_lock();
-	sock = rcu_dereference(wg->sock6);
+	rcu_read_lock_bh();
+	sock = rcu_dereference_bh(wg->sock6);
 
 	if (unlikely(!sock)) {
 		ret = -ENONET;
@@ -116,29 +121,25 @@ static inline int send6(struct wireguard_device *wg, struct sk_buff *skb, struct
 		}
 		ret = ipv6_stub->ipv6_dst_lookup(sock_net(sock), sock, &dst, &fl);
 		if (unlikely(ret)) {
-			net_dbg_ratelimited("No route to %pISpfsc, error %d\n", &endpoint->addr, ret);
+			net_dbg_ratelimited("%s: No route to %pISpfsc, error %d\n", wg->dev->name, &endpoint->addr, ret);
 			goto err;
 		} else if (unlikely(dst->dev == skb->dev)) {
 			dst_release(dst);
 			ret = -ELOOP;
-			net_dbg_ratelimited("Avoiding routing loop to %pISpfsc\n", &endpoint->addr);
+			net_dbg_ratelimited("%s: Avoiding routing loop to %pISpfsc\n", wg->dev->name, &endpoint->addr);
 			goto err;
 		}
 		if (cache)
 			dst_cache_set_ip6(cache, dst, &fl.saddr);
 	}
 
-	udp_tunnel6_xmit_skb(dst, sock, skb, skb->dev,
-			     &fl.saddr, &fl.daddr,
-			     ds, ip6_dst_hoplimit(dst), 0,
-			     fl.fl6_sport, fl.fl6_dport,
-			     false);
+	udp_tunnel6_xmit_skb(dst, sock, skb, skb->dev, &fl.saddr, &fl.daddr, ds, ip6_dst_hoplimit(dst), 0, fl.fl6_sport, fl.fl6_dport, false);
 	goto out;
 
 err:
 	kfree_skb(skb);
 out:
-	rcu_read_unlock();
+	rcu_read_unlock_bh();
 	return ret;
 #else
 	return -EAFNOSUPPORT;
@@ -165,8 +166,10 @@ int socket_send_skb_to_peer(struct wireguard_peer *peer, struct sk_buff *skb, u8
 int socket_send_buffer_to_peer(struct wireguard_peer *peer, void *buffer, size_t len, u8 ds)
 {
 	struct sk_buff *skb = alloc_skb(len + SKB_HEADER_LEN, GFP_ATOMIC);
+
 	if (unlikely(!skb))
 		return -ENOMEM;
+
 	skb_reserve(skb, SKB_HEADER_LEN);
 	memcpy(skb_put(skb, len), buffer, len);
 	return socket_send_skb_to_peer(peer, skb, ds);
@@ -203,12 +206,13 @@ int socket_send_buffer_as_reply_to_skb(struct wireguard_device *wg, struct sk_bu
 int socket_endpoint_from_skb(struct endpoint *endpoint, struct sk_buff *skb)
 {
 	memset(endpoint, 0, sizeof(struct endpoint));
-	if (ip_hdr(skb)->version == 4) {
+	if (skb->protocol == htons(ETH_P_IP)) {
 		endpoint->addr4.sin_family = AF_INET;
 		endpoint->addr4.sin_port = udp_hdr(skb)->source;
 		endpoint->addr4.sin_addr.s_addr = ip_hdr(skb)->saddr;
 		endpoint->src4.s_addr = ip_hdr(skb)->daddr;
-	} else if (ip_hdr(skb)->version == 6) {
+		endpoint->src_if4 = skb->skb_iif;
+	} else if (skb->protocol == htons(ETH_P_IPV6)) {
 		endpoint->addr6.sin6_family = AF_INET6;
 		endpoint->addr6.sin6_port = udp_hdr(skb)->source;
 		endpoint->addr6.sin6_addr = ipv6_hdr(skb)->saddr;
@@ -227,12 +231,14 @@ void socket_set_peer_endpoint(struct wireguard_peer *peer, struct endpoint *endp
 		if (likely(peer->endpoint.addr4.sin_family == AF_INET &&
 			   peer->endpoint.addr4.sin_port == endpoint->addr4.sin_port &&
 			   peer->endpoint.addr4.sin_addr.s_addr == endpoint->addr4.sin_addr.s_addr &&
-			   peer->endpoint.src4.s_addr == endpoint->src4.s_addr))
+			   peer->endpoint.src4.s_addr == endpoint->src4.s_addr &&
+			   peer->endpoint.src_if4 == endpoint->src_if4))
 			goto out;
 		read_unlock_bh(&peer->endpoint_lock);
 		write_lock_bh(&peer->endpoint_lock);
 		peer->endpoint.addr4 = endpoint->addr4;
 		peer->endpoint.src4 = endpoint->src4;
+		peer->endpoint.src_if4 = endpoint->src_if4;
 	} else if (endpoint->addr.sa_family == AF_INET6) {
 		read_lock_bh(&peer->endpoint_lock);
 		if (likely(peer->endpoint.addr6.sin6_family == AF_INET6 &&
@@ -280,45 +286,6 @@ err:
 	return 0;
 }
 
-/* Generates a default port from the interface name.
- * wg0 --> 51820
- * wg1 --> 51821
- * wg2 --> 51822
- * wg100 --> 51920
- * wg60000 --> 46285
- * blahbla --> 51820
- * 50 --> 51870
- */
-static u16 generate_default_incoming_port(struct wireguard_device *wg)
-{
-	u16 port = 51820;
-	unsigned long parsed;
-	char *name, *digit_begin;
-	size_t len;
-
-	ASSERT_RTNL();
-
-	name = netdev_pub(wg)->name;
-	len = strlen(name);
-	if (!len)
-		return port;
-	digit_begin = name + len - 1;
-	while (digit_begin >= name) {
-		if (isdigit(*digit_begin))
-			--digit_begin;
-		else
-			break;
-	}
-	++digit_begin;
-	if (!*digit_begin)
-		return port;
-	if (!kstrtoul(digit_begin, 10, &parsed))
-		port += parsed;
-	if (!port)
-		++port;
-	return port;
-}
-
 static inline void sock_free(struct sock *sock)
 {
 	if (unlikely(!sock))
@@ -337,60 +304,63 @@ static inline void set_sock_opts(struct socket *sock)
 int socket_init(struct wireguard_device *wg)
 {
 	int ret = 0;
-	struct socket *new4 = NULL;
-	struct udp_port_cfg port4 = {
-		.family = AF_INET,
-		.local_ip.s_addr = htonl(INADDR_ANY),
-		.use_udp_checksums = true
-	};
-#if IS_ENABLED(CONFIG_IPV6)
-	struct socket *new6 = NULL;
-	struct udp_port_cfg port6 = {
-		.family = AF_INET6,
-		.local_ip6 = IN6ADDR_ANY_INIT,
-		.use_udp6_tx_checksums = true,
-		.use_udp6_rx_checksums = true,
-		.ipv6_v6only = true
-	};
-#endif
 	struct udp_tunnel_sock_cfg cfg = {
 		.sk_user_data = wg,
 		.encap_type = 1,
 		.encap_rcv = receive
 	};
+	struct socket *new4 = NULL;
+	struct udp_port_cfg port4 = {
+		.family = AF_INET,
+		.local_ip.s_addr = htonl(INADDR_ANY),
+		.local_udp_port = htons(wg->incoming_port),
+		.use_udp_checksums = true
+	};
+#if IS_ENABLED(CONFIG_IPV6)
+	int retries = 0;
+	struct socket *new6 = NULL;
+	struct udp_port_cfg port6 = {
+		.family = AF_INET6,
+		.local_ip6 = IN6ADDR_ANY_INIT,
+		.local_udp_port = htons(wg->incoming_port),
+		.use_udp6_tx_checksums = true,
+		.use_udp6_rx_checksums = true,
+		.ipv6_v6only = true
+	};
+#endif
 
 	mutex_lock(&wg->socket_update_lock);
-
-	if (rcu_dereference_protected(wg->sock4, lockdep_is_held(&wg->socket_update_lock)) ||
-	    rcu_dereference_protected(wg->sock6, lockdep_is_held(&wg->socket_update_lock))) {
+#if IS_ENABLED(CONFIG_IPV6)
+retry:
+#endif
+	if (rcu_dereference_protected(wg->sock4, lockdep_is_held(&wg->socket_update_lock)) || rcu_dereference_protected(wg->sock6, lockdep_is_held(&wg->socket_update_lock))) {
 		ret = -EADDRINUSE;
 		goto out;
 	}
 
-	if (!wg->incoming_port)
-		wg->incoming_port = generate_default_incoming_port(wg);
-	port4.local_udp_port =
-#if IS_ENABLED(CONFIG_IPV6)
-		port6.local_udp_port =
-#endif
-		htons(wg->incoming_port);
-
 	ret = udp_sock_create(wg->creating_net, &port4, &new4);
 	if (ret < 0) {
-		pr_err("Could not create IPv4 socket\n");
+		pr_err("%s: Could not create IPv4 socket\n", wg->dev->name);
 		goto out;
 	}
-
+	wg->incoming_port = ntohs(inet_sk(new4->sk)->inet_sport);
 	set_sock_opts(new4);
 	setup_udp_tunnel_sock(wg->creating_net, new4, &cfg);
 	rcu_assign_pointer(wg->sock4, new4->sk);
 
 #if IS_ENABLED(CONFIG_IPV6)
+	if (!ipv6_mod_enabled())
+		goto out;
+	port6.local_udp_port = htons(wg->incoming_port);
 	ret = udp_sock_create(wg->creating_net, &port6, &new6);
 	if (ret < 0) {
-		pr_err("Could not create IPv6 socket\n");
 		udp_tunnel_sock_release(new4);
 		rcu_assign_pointer(wg->sock4, NULL);
+		if (ret == -EADDRINUSE && !port4.local_udp_port && retries++ < 100)
+			goto retry;
+		if (!port4.local_udp_port)
+			wg->incoming_port = 0;
+		pr_err("%s: Could not create IPv6 socket\n", wg->dev->name);
 		goto out;
 	}
 	set_sock_opts(new6);
@@ -406,13 +376,15 @@ out:
 void socket_uninit(struct wireguard_device *wg)
 {
 	struct sock *old4, *old6;
+
 	mutex_lock(&wg->socket_update_lock);
 	old4 = rcu_dereference_protected(wg->sock4, lockdep_is_held(&wg->socket_update_lock));
 	old6 = rcu_dereference_protected(wg->sock6, lockdep_is_held(&wg->socket_update_lock));
 	rcu_assign_pointer(wg->sock4, NULL);
 	rcu_assign_pointer(wg->sock6, NULL);
 	mutex_unlock(&wg->socket_update_lock);
-	synchronize_rcu();
+	synchronize_rcu_bh();
+	synchronize_net();
 	sock_free(old4);
 	sock_free(old6);
 }
